from statemachine import StateMachine, State

from pkg_trainmote.actions.actionInterface import ActionInterface
from .models.Program import Program
from .models.Action import Action
from typing import Optional
from .actions import actionHelper

class ProgramMachine(StateMachine):

    __program: Optional[Program]
    __action_index: int = 0
    __actionInterface: Optional[ActionInterface] = None

    isRunning: bool = False

    def start(self, program: Program):

        if self.is_readyForAction:
            self.__program = program
            self.startProgram()

    readyForAction = State('readyForAction', initial=True)
    runningAction = State('RunningAction')
    actionFinished = State('ActionFinished')
    preparingAction = State('preparingAction')

    startProgram = readyForAction.to(preparingAction)
    cancelProgram = runningAction.to(readyForAction)
    endAction = runningAction.to(actionFinished)
    prepareForAction = actionFinished.to(preparingAction)
    startAction = preparingAction.to(runningAction)
    endProgram = preparingAction.to(readyForAction)

    def on_startProgram(self):
        print('startProgram')
        self.isRunning = True

    def on_enter_runningAction(self):
        print('startAction')

        def actionCallback():
            print("action finished")
            self.endAction()
        self.__actionInterface.runAction(actionCallback)

    def on_cancelProgram(self):
        print('cancelProgram')
        self.isRunning = False
        self.__program = None

    def on_endAction(self):
        print('endAction')
        self.__action_index = self.__action_index + 1

    def on_enter_actionFinished(self):
        self.prepareForAction()

    def on_enter_preparingAction(self):
        print('preparing')
        self.prepare()

    def prepare(self):
        if self.__action_index < len(self.__program.actions):
            action = self.__program.actions[self.__action_index]
            self.__actionInterface = actionHelper.getProgramAction(action)
            if self.__actionInterface is not None:
                self.startAction()
                return
        self.endProgram()

    def on_endProgram(self):
        print('on_endProgram')
        self.isRunning = False
        self.__program = None



from typing import Optional, List, Dict, Any
from .Action import Action
from types import SimpleNamespace
import json
from pkg_trainmote.validators.validator import Validator

class Program():

    def __init__(
        self,
        uid: Optional[str],
        actions: List[Action],
        name: Optional[str]
    ):
        self.uid = uid
        self.actions = actions
        self.name = name

    def to_dict(self):
        return {
            "uid": self.uid,
            "actions": [action.to_dict() for action in self.actions],
            "name": self.name
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any], id: str):
        actions: List[Action] = []
        for action in data.get("actions"):
            actions.append(Action.from_dict(action))

        program = cls(
            id,
            actions,
            str(data.get("name"))
        )
        return program

    @classmethod
    def from_Json(cls, data: Any):
        if Validator().validateDict(data, "program_scheme") is True:
            mProgram = json.loads(json.dumps(data), object_hook=lambda d: SimpleNamespace(**d))
            if hasattr(mProgram, 'uid'): 
                return cls(mProgram.uid, mProgram.actions, mProgram.name)
            else:
                return cls(None, mProgram.actions, mProgram.name)
        else:
            raise ValueError("Invalid json")
    


from pkg_trainmote.databaseControllerModule import DatabaseController
from typing import List
from pkg_trainmote.models.GPIORelaisModel import GPIORelaisModel

def containsPin(relais_id: int, relais: List[GPIORelaisModel]) -> bool:
    for rel in relais:
        if rel.relais_id == relais_id:
            return True
    return False

def isAlreadyInUse(pin: int):
    database = DatabaseController()
    stops = database.getAllStopModels()
    switchs = database.getAllSwichtModels()
    relaisIsStop = containsPin(pin, stops)
    if relaisIsStop:
        raise ValueError("Pin is already in use as stop point")
    relaisIsSwitch = containsPin(pin, switchs)
    if relaisIsSwitch:
        raise ValueError("Pin is already in use as switch")
    config = database.getConfig()
    if config is not None:
        if config.powerRelais is pin:
            raise ValueError("Pin is already in use as power relais")
        if config.switchPowerRelais is pin:
            raise ValueError("Pin is already in use as switch power relais")
        if config.stateRelais is pin:
            raise ValueError("Pin is already in use as state relais")



import json
import RPi.GPIO as GPIO
from .traintrackingservice import TrackingService
from .models.GPIORelaisModel import GPIORelaisAdapter, GPIORelaisModel, GPIOSwitchHelper
from .models.GPIORelaisModel import GPIOStoppingPoint
from .models.GPIORelaisModel import GPIOSwitchPoint
from .databaseControllerModule import DatabaseController
from typing import Optional

gpioRelais = []
trackingServices = []
validGpios = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27]
switchPowerRelais: Optional[GPIORelaisModel] = None
# Inital Loading and Setup


def setup():
    print("GPIOService setup")
    try:
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(True)
    except Exception as identifier:
        print(identifier)
    loadInitialData()
    setupTrackingDefault()


def loadInitialData():
    global switchPowerRelais
    config = DatabaseController().getConfig()
    if config is not None:
        if config.switchPowerRelais is not None:
            # Initialise GPIO for switch power relais
            switchPowerRelais = GPIORelaisModel("switchPowerRelais", config.switchPowerRelais)
            GPIO.setup(switchPowerRelais.relais_id, GPIO.OUT, initial=GPIO.HIGH)

    switchModels = DatabaseController().getAllSwichtModels()
    for model in switchModels:
        addSwitch(model)

    stopModels = DatabaseController().getAllStopModels()
    for stop in stopModels:
        addRelais(stop)

def removeRelais(relais: GPIORelaisModel):
    if relais in gpioRelais:
        gpioRelais.remove(relais)
        GPIO.cleanup(relais.relais_id)

def addSwitch(switch: GPIOSwitchPoint):    
    if switch.needsPowerOn and switchPowerRelais is not None:
        switch.setPowerRelais(switchPowerRelais)
    addRelais(switch)

def addRelais(relais: GPIORelaisModel):
    try:
        GPIO.setup(relais.relais_id, GPIO.OUT, initial=relais.defaultValue)
        gpioRelais.append(relais)
    except Exception as e:
        print(e)

def setupTrackingDefault():
    for relais in gpioRelais:
        if isinstance(relais, GPIOStoppingPoint) and relais.mess_id is not None:
            startTrackingFor(relais)


def startTrackingFor(relais):
    trackingService = TrackingService(relais)
    trackingServices.append(trackingService)
    trackingService.startTracking()


def resetData():
    DatabaseController().removeAll()
    del gpioRelais[:]
    del trackingServices[:]

def clean():
    print("Clean GPIOs")
    GPIO.cleanup()

def getValueForPin(pin):
    return GPIO.input(pin)


def getSwitchWithID(id):
    for relais in gpioRelais:
        if isinstance(relais, GPIOSwitchPoint):
            if relais.uid == id:
                return relais
    return None

def getStopWithID(id):
    for relais in gpioRelais:
        if isinstance(relais, GPIOStoppingPoint):
            if relais.uid == id:
                return relais
    return None

# Relais Actions


def switchPin(relais):
    if relais.getStatus():
        if isinstance(relais, GPIOStoppingPoint):
            trackingService = next(
                (tracker for tracker in trackingServices if tracker.stoppingPoint.uid == relais.uid),
                None
            )
            if trackingService:
                trackingService.stopTracking()
                trackingServices.remove(trackingService)
        return relais.setStatus(GPIO.LOW)
    else:
        if isinstance(relais, GPIOStoppingPoint) and relais.mess_id is not None:
            startTrackingFor(relais)
        return relais.setStatus(GPIO.HIGH)


def receivedMessage(message):
    return "msg:Not valid json"

##
# Switch
##


def getSwitch(id: str) -> str:
    switch = getSwitchFor(id)
    if switch is not None:
        return json.dumps(switch.to_dict())
    raise ValueError("Switch for id {} not found".format(id))


def getAllSwitches():
    return json.dumps([ob.to_dict() for ob in DatabaseController().getAllSwichtModels()])


def setSwitch(id: str) -> str:
    relais = getSwitchWithID(id)
    if relais is not None:
        switch = getSwitchFor(id)
        if switch is not None:
            newValue = switchPin(relais)
            return json.dumps({"model": switch.to_dict(), "currentValue": newValue})
    raise ValueError("Relais not found for id {}".format(id))

def getSwitchFor(uid: str) -> Optional[GPIOSwitchPoint]:
    for switch in DatabaseController().getAllSwichtModels():
        if switch.uid == uid:
            return switch
    return None

def createSwitch(data):
    gpioRelais = GPIORelaisAdapter.getGPIORelaisFor(data)
    if gpioRelais is not None:
        switchType = data.get("switchType")
        needsPowerOn = data.get("needsPowerOn")
        if GPIOSwitchHelper.isValidType(switchType):
            switch = GPIOSwitchPoint.fromParent(gpioRelais, switchType, needsPowerOn)
            result = storeSwitch(switch)
            if result is not None and result.relais_id is not None:
                return json.dumps(result.to_dict())
            else:
                raise ValueError("Could not create switch")
        else:
            raise ValueError("Inavlid switch type")
    else:
        raise ValueError("Could not create switch")

# Stores a Switch Point to the database if is a valid pin number.
# Throws error if not valid. Adds the relais to the gpioservice relais.
def storeSwitch(model: GPIOSwitchPoint) -> Optional[GPIOSwitchPoint]:
    if (model.relais_id is not None and isValidRaspberryPiGPIO(model.relais_id)):
        databaseController = DatabaseController()
        result = databaseController.insertSwitchModel(
            model.relais_id,
            model.switchType,
            model.needsPowerOn,
            model.defaultValue,
            model.name,
            model.description
        )
        if (result is not None):
            switch = databaseController.getSwitch(result)
            if (switch is not None):
                switch.setDefaultValue(model.defaultValue)
                addRelais(switch)
                return switch
    else:
        raise ValueError("Invalid gpio")


##
# Stop Point
##


def getStop(id: str):
    stop = getStopFor(id)
    if stop is not None:
        return json.dumps(stop.to_dict())
    return json.dumps({"error": "Stop for id {} not found".format(id)})


def getAllStopPoints():
    return json.dumps([ob.to_dict() for ob in DatabaseController().getAllStopModels()])

def setStop(id: str):
    relais = getStopWithID(id)
    if relais is not None:
        stop = getStopFor(id)
        if stop is not None:
            newValue = switchPin(relais)
            return json.dumps({"model": stop.to_dict(), "currentValue": newValue})
    raise ValueError("Relais not found for id {}".format(id))

def getStopFor(uid: str) -> Optional[GPIOStoppingPoint]:
    for stop in DatabaseController().getAllStopModels():
        if stop.uid == uid:
            return stop
    return None

def createStop(data):
    gpioRelais = GPIORelaisAdapter.getGPIORelaisFor(data)
    if gpioRelais is not None:
        mess_id = int(data.get("mess_id"))
        stop = GPIOStoppingPoint.fromParent(gpioRelais, mess_id)
        result = storeStop(stop)
        if result is not None and result.relais_id is not None:
            return json.dumps(result.to_dict())
        else:
            raise ValueError("Could not create stop")
    else:
        raise ValueError("Could not create stop")

# Stores a Stop Point to the database if is a valid pin number.
# Throws error if not valid. Adds the relais to the gpioservice relais.
def storeStop(model: GPIOStoppingPoint) -> Optional[GPIOStoppingPoint]:
    if (model.relais_id is not None and isValidRaspberryPiGPIO(model.relais_id)):
        databaseController = DatabaseController()
        result = databaseController.insertStopModel(
            model.relais_id,
            model.mess_id,
            model.name,
            model.defaultValue,
            model.description
        )
        if (result is not None):
            stop = databaseController.getStop(result)
            if (stop is not None):
                addRelais(stop)
                return stop
    else:
        raise ValueError("Invalid gpio")

def setAllToDefault():
    for relais in gpioRelais:
        relais.toDefault()

# Validation

def isValidRaspberryPiGPIO(pinNumber: int):
    return pinNumber in validGpios



import threading
from typing import Optional
from pkg_trainmote.models.Action import Action
from pkg_trainmote.actions.actionInterface import ActionInterface
import time

class TimerThread(threading.Thread):

    seconds = 0

    def __init__(self, interval: int, callback):
        threading.Thread.__init__(self)
        self.seconds = interval
        self.kill = threading.Event()
        self.callback = callback

    def run(self):
        time.sleep(self.seconds)
        if self.callback is not None:
            self.callback()

class TimerAction(ActionInterface):

    __action: Optional[Action] = None
    timer: Optional[TimerThread] = None
    __seconds: int = 0

    def __init__(self, action: Action) -> None:
        self.__action = action

    def prepareAction(self):
        if self.__action is not None:
            hhSeconds = int(self.__action.values[0]) * 3600
            mmSeconds = int(self.__action.values[1]) * 60
            self.__seconds = + hhSeconds + mmSeconds + int(self.__action.values[2])
            print(self.__seconds)

    def runAction(self, _callback):
        TimerThread(self.__seconds, _callback)

    def cancelAction(self):
        self.timer.join()
        self.timer.kill.set()
        pass


# class BaseThread(threading.Thread):
#     def __init__(self, callback=None, callback_args=None, *args, **kwargs):
#         target = kwargs.pop('target')
#         super(BaseThread, self).__init__(target=self.target_with_callback, *args, **kwargs)
#         self.callback = callback
#         self.method = target
#         self.callback_args = callback_args

#     def target_with_callback(self):
#         self.method()
#         if self.callback is not None:
#             self.callback(*self.callback_args)



from typing import Optional, Dict, Any, List
import enum

class Action():

    def __init__(
        self,
        uid: Optional[str],
        type: str,
        position: int,
        values: List[str],
        name: Optional[str]
    ):
        self.uid = uid
        self.type = type
        self.position = position
        self.name = name
        self.values = values

    def to_dict(self):
        return {
            "uid": self.uid,
            "type": self.type,
            "position": self.position,
            "values": self.values,
            "name": self.name
        }    

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        values: List[str] = []
        for value in data.get("values"):
            values.append(value)
        action = cls(
            str(data.get("uid")),
            str(data.get("type")),
            int(str(data.get("position"))),
            values,
            data.get("name")
        )
        return action


class ActionType(enum.Enum):
    TM_SET_SWITCH = "SetSwitch"
    TM_SET_STOP = "SetStop"
    TM_PERFORM_TIMER = "WaitFor"
    TM_PERFORM_ALARM = "StartAtClock"



from pkg_trainmote.stateControllerModule import StateController
from . import gpioservice
from flask import Flask
from .powerControllerModule import PowerThread
from .configControllerModule import ConfigController
from . import stateControllerModule
from .libInstaller import LibInstaller
from .databaseControllerModule import DatabaseController
from .stopPointApiController import stopPointApi
from .deviceApiController import deviceApiBlueprint
from .switchApiController import switchApiBlueprint
from .configApiController import configApi
from .programApiController import programApi
from .authentication import auth
from typing import Optional
import sys
import os
import json
import signal
import socket

stateController: Optional[StateController]
dataBaseController: Optional[DatabaseController]
powerThread: Optional[PowerThread]
config: Optional[ConfigController]
app = Flask(__name__)
app.register_blueprint(stopPointApi)
app.register_blueprint(deviceApiBlueprint)
app.register_blueprint(switchApiBlueprint)
app.register_blueprint(configApi)
app.register_blueprint(programApi)

# users = {
#    "guest": generate_password_hash("S5Va4BUzjj4K")
# }

mVersion: Optional[str] = None

def loadPersistentData():
    if config.loadPreferences():
        if not config.isSQLiteInstalled():
            libInstaller = LibInstaller()
            libInstaller.installSQLite()
            if config.setSQLiteInstalled():
                restart()
            else:
                shutDown()


def setup(version):
    global mVersion
    mVersion = version
    gpioservice.setup()
    global dataBaseController
    dataBaseController = DatabaseController()
    dataBaseController.checkUpdate(version)

    global powerThread
    powerThread = None

    global stateController
    stateController = None

    conf = DatabaseController().getConfig()
    if conf is not None:
        if conf.powerRelais is not None:
            setupPowerGPIO(conf.powerRelais)
        if conf.stateRelais is not None:
            stateController = stateControllerModule.StateController(conf.stateRelais)
            stateController.setState(stateControllerModule.STATE_NOT_CONNECTED)

    global config
    config = ConfigController()
    print("Start webserver")
    app.run(host="0.0.0.0")
    signal.signal(signal.SIGINT, handler)

##
# Setup PowerThread to track user event to shut down.
##
def setupPowerGPIO(pin: int):
    powerThread = PowerThread(pin)
    powerThread.start()

@app.route('/trainmote/api/v1')
def hello_world():
    if stateController is not None:
        stateController.setState(stateControllerModule.STATE_CONNECTED)
    conf = dataBaseController.getConfig()
    if conf is not None:
        return json.dumps({
            "trainmote": "trainmote.module.felix-nievelstein.de",
            "version": mVersion,
            "name": conf.deviceName
        })
    else:
        return json.dumps({
            "trainmote": "trainmote.module.felix-nievelstein.de",
            "version": mVersion,
            "name": socket.gethostname()
        })


def restart():
    shutDown()
    os.execv(sys.executable, ['python'] + sys.argv)


def shutDown():
    print("Server going down")
    gpioservice.clean()


def closeClientConnection():
    print("Closing client socket")


def handler(signal, frame):
    shutDown()
    sys.exit(0)

def stopRunningThreads():
    if powerThread is not None:
        if powerThread.is_alive():
            powerThread.stop()
            powerThread.isTurningOff = True
            powerThread.join()
    if stateController is not None:
        stateController.setState(stateControllerModule.STATE_SHUTDOWN)
        stateController.stop()



from pkg_trainmote.actions.actionInterface import ActionInterface
from pkg_trainmote.models.Action import Action, ActionType
from pkg_trainmote.actions.setSwitchAction import SetSwitchAction
from pkg_trainmote.actions.setStopAction import SetStopAction
from pkg_trainmote.actions.timerAction import TimerAction
from typing import Optional

def getProgramAction(action: Action) -> Optional[ActionInterface]:
    if action.type == ActionType.TM_SET_SWITCH.value:
        return SetSwitchAction(action)
    elif action.type == ActionType.TM_SET_STOP.value:
        return SetStopAction(action)
    elif action.type == ActionType.TM_PERFORM_TIMER:
        return TimerAction(action)
    return None
